'
'*******************************************************************************
'
' Description: VB Module for accessing TWAIN compatible scanner (VB 5, 6)
'
'------------- >>> Converted to PBWIN 8.x <<< -------------
'------------- >>> by Babakhanov UserName    <<< -------------
'
'
' Author:      Lumir Mik (lmik@seznam.cz)
'
' Version:     1.0
'
' License:     Free to any use. If you change some part of this code, please,
'              mention it here.
'              Receive it as my contribution to free programmer sources
'              in which I found much help and inspiration.
'
' There are 3 public functions in this module:
'
'   1. PopupSelectSourceDialog
'           shows TWAIN dialog for selecting default source for acquisition
'
'   2. TransferWithoutUI
'           transfers one image from TWAIN data source without showing
'           the data source user interface (silent transfer). The programmer
'           can set following attributes of the image:
'               - resolution (DPI)
'               - colour depth - monochromatic, grey, fullcolour
'               - image size and position on the scanner glass
'                       - left, top, right, bottom (in inches).
'           The image is saved into the BMP file.
'
'   3. TransferWithUI
'           transfers one image from TWAIN data source using the data
'           source user interface to set image attributes.
'           The image is saved into the BMP file.
'
'*******************************************************************************

#DIM ALL

'-------------------------
' Declaration for WIN32API
'-------------------------
#IF NOT %DEF(%WIN32API)
   #INCLUDE "WIN32API.INC"
#ENDIF


'-----------------------------
' Declaration for TWAIN_32.DLL
'-----------------------------
DECLARE FUNCTION DSM_Entry LIB "Twain_32.dll" ALIAS "DSM_Entry"_
                                   (BYREF pOrigin AS ANY, _
                                    BYREF pDest AS ANY, _
                                    BYVAL DG  AS DWORD, _
                                    BYVAL DAT AS WORD, _
                                    BYVAL MSG AS WORD, _
                                    BYREF pData AS ANY) AS INTEGER

TYPE TW_VERSION
    MajorNum        AS INTEGER                  ' TW_UINT16
    MinorNum        AS INTEGER                  ' TW_UINT16
    Language        AS INTEGER                  ' TW_UINT16
    Country         AS INTEGER                  ' TW_UINT16
    Info(1 TO 34)   AS BYTE                     ' TW_STR32
END TYPE

TYPE TW_IDENTITY
    Id                      AS LONG             ' TW_UINT32
    Version                 AS TW_VERSION       ' TW_VERSION
    ProtocolMajor           AS INTEGER          ' TW_UINT16
    ProtocolMinor           AS INTEGER          ' TW_UINT16
    SupportedGroups1        AS INTEGER          ' TW_UINT32
    SupportedGroups2        AS INTEGER
    Manufacturer(1 TO 34)   AS BYTE             ' TW_STR32
    ProductFamily(1 TO 34)  AS BYTE             ' TW_STR32
    ProductName(1 TO 34)    AS BYTE             ' TW_STR32
END TYPE

TYPE TW_USERINTERFACE
    ShowUI   AS INTEGER                         ' TW_BOOL
    ModalUI  AS INTEGER                         ' TW_BOOL
    hParent  AS LONG                            ' TW_HANDLE
END TYPE

TYPE TW_PENDINGXFERS
    Count       AS INTEGER                      ' TW_UINT16
    Reserved1   AS INTEGER                      ' TW_UINT32
    Reserved2   AS INTEGER
END TYPE

TYPE TW_ONEVALUE
    ItemType AS INTEGER                         ' TW_UINT16
    Item1    AS INTEGER                         ' TW_UINT32
    Item2    AS INTEGER
END TYPE

TYPE TW_CAPABILITY
    Cap          AS INTEGER                     ' TW_UINT16
    ConType      AS INTEGER                     ' TW_UINT16
    hContainer   AS LONG                        ' TW_HANDLE
END TYPE

TYPE TW_FIX32
    Whole   AS INTEGER                          ' TW_INT16
    FRAC    AS INTEGER                          ' TW_UINT16
END TYPE

TYPE TW_FRAME
    LEFT     AS TW_FIX32                        ' TW_FIX32
    Top      AS TW_FIX32                        ' TW_FIX32
    RIGHT    AS TW_FIX32                        ' TW_FIX32
    Bottom   AS TW_FIX32                        ' TW_FIX32
END TYPE

TYPE TW_IMAGELAYOUT
    FRAME            AS TW_FRAME                ' TW_FRAME
    DocumentNumber   AS LONG                    ' TW_UINT32
    PageNumber       AS LONG                    ' TW_UINT32
    FrameNumber      AS LONG                    ' TW_UINT32
END TYPE

TYPE TW_EVENT
    pEvent      AS LONG                         ' TW_MEMREF
    TWMessage   AS INTEGER                      ' TW_UINT16
END TYPE

%DG_CONTROL = 1
%DG_IMAGE = 2

%MSG_GET = 1
%MSG_SET = 6
%MSG_XFERREADY = 257
%MSG_CLOSEDSREQ = 258
%MSG_OPENDSM = 769
%MSG_CLOSEDSM = 770
%MSG_OPENDS = 1025
%MSG_CLOSEDS = 1026
%MSG_USERSELECT = 1027
%MSG_DISABLEDS = 1281
%MSG_ENABLEDS = 1282
%MSG_PROCESSEVENT = 1537
%MSG_ENDXFER = 1793

%DAT_CAPABILITY = 1
%DAT_EVENT = 2
%DAT_IDENTITY = 3
%DAT_PARENT = 4
%DAT_PENDINGXFERS = 5
%DAT_USERINTERFACE = 9
%DAT_IMAGELAYOUT = 258
%DAT_IMAGENATIVEXFER = 260

%TWRC_SUCCESS = 0
%TWRC_CHECKSTATUS = 2
%TWRC_DSEVENT = 4
%TWRC_NOTDSEVENT = 5
%TWRC_XFERDONE = 6

%TWLG_CZECH = 45

%TWCY_CZECHOSLOVAKIA = 42

%TWON_PROTOCOLMAJOR = 1
%TWON_ONEVALUE = 5
%TWON_PROTOCOLMINOR = 9

TYPE MSG
    hwnd    AS LONG
    message AS LONG
    wParam  AS LONG
    lParam  AS LONG
    time    AS LONG
    pt      AS POINTAPI
END TYPE

%GHND = 66

'----------------------------
' Declaration for this Module
'----------------------------
GLOBAL m_tAppID     AS TW_IDENTITY, _
       m_tSrcID     AS TW_IDENTITY, _
       m_lHndMsgWin AS DWORD

' Enum TWAIN_MDL_COLOURTYPE
%TWPT_BW =  0
%TWPT_GRAY = 1
%TWPT_RGB = 2

' Enum TWAIN_MDL_CAPABILITY
%XFERCOUNT = 1                       ' CAP_XFERCOUNT
%PIXELTYPE = 257                     ' ICAP_PIXELTYPE
%INDICATORS = 4107                   ' CAP_INDICATORS
%UICONTROLLABLE = 4110               ' CAP_UICONTROLLABLE
%ICAP_PHYSICALWIDTH = 4369                ' ICAP_PSYSICALWIDTH
%ICAP_PHYSICALHEIGHT = 4370               ' ICAP_PSYSICALHEIGHT
%XRESOLUTION = 4376                  ' ICAP_XRESOLUTION
%YRESOLUTION = 4377                  ' ICAP_YRESOLUTION
%BITDEPTH = 4395                     ' ICAP_BITDEPTH

' Enum TWAIN_MDL_ITEMYPE
%INT16 = 1                           ' TW_INT16      short
%UINT16 = 4                          ' TW_UINT16     unsigned short
%BOOL = 6                            ' TW_BOOL       unsigned short
%FIX32 = 7                           ' TW_FIX32      structure

DECLARE FUNCTION CloseTwainDS() AS LONG
DECLARE FUNCTION TransferWithUI(BYVAL sBMPFileName AS STRING) AS LONG
DECLARE FUNCTION CloseTwainDSM() AS LONG
DECLARE FUNCTION Fix32ToFloat(BYREF tFix32 AS TW_FIX32) AS SINGLE
DECLARE FUNCTION FloatToFix32(BYREF sngSrc AS SINGLE, BYREF tFix32 AS TW_FIX32) AS LONG
DECLARE FUNCTION FromUnsignedLong(BYREF lSrc AS LONG) AS SINGLE
DECLARE FUNCTION FromUnsignedShort(BYREF iSrc AS INTEGER) AS LONG
DECLARE FUNCTION OpenTwainDS() AS LONG
DECLARE FUNCTION OpenTwainDSM() AS LONG
DECLARE FUNCTION PopupSelectSourceDialog() AS LONG
DECLARE FUNCTION SaveDIBToFile(BYREF lhDIB AS LONG, BYREF sFileName AS STRING) AS LONG
DECLARE FUNCTION SetImageSize(BYREF sngLeft AS SINGLE, _
                              BYREF sngTop AS SINGLE, BYREF sngRight AS SINGLE,BYREF sngBottom AS SINGLE) AS LONG
DECLARE FUNCTION ToUnsignedLong(BYREF sngSrc AS SINGLE, BYREF lDst AS LONG) AS LONG
DECLARE FUNCTION ToUnsignedShort(BYREF lSrc AS LONG, BYREF iDst AS INTEGER) AS LONG
DECLARE FUNCTION TransferWithoutUI(BYVAL sngResolution AS SINGLE, _
                                  BYVAL tColourType AS LONG, _
                                  BYVAL sngImageLeft AS SINGLE, _
                                  BYVAL sngImageTop AS SINGLE, _
                                  BYVAL sngImageRight AS SINGLE, _
                                  BYVAL sngImageBottom AS SINGLE, _
                                  BYVAL sBMPFileName AS STRING) AS LONG
DECLARE FUNCTION TransferWithUI(BYVAL sBMPFileName AS STRING) AS LONG
DECLARE FUNCTION TwainGetOneValue(BYVAL Cap AS LONG, BYREF Item AS VARIANT) AS LONG
DECLARE FUNCTION TwainSetOneValue(BYVAL Cap AS LONG, BYVAL ItemType AS LONG, _ 'TWAIN_MDL_ITEMYPE, _
                                  BYREF Item AS VARIANT) AS LONG
DECLARE FUNCTION TwainTransfer(BYREF blShowUI AS INTEGER, BYREF lDIBHandle AS LONG) AS LONG

FUNCTION TransferWithoutUI(BYVAL sngResolution AS SINGLE, _
                                  BYVAL tColourType AS LONG, _
                                  BYVAL sngImageLeft AS SINGLE, _
                                  BYVAL sngImageTop AS SINGLE, _
                                  BYVAL sngImageRight AS SINGLE, _
                                  BYVAL sngImageBottom AS SINGLE, _
                                  BYVAL sBMPFileName AS STRING) AS LONG

    '----------------------------------------------------------------------------
    ' Function transfers one image from Twain data source without showing
    '   the data source user interface (silent transfer).
    '
    ' Input values
    '   - sngResolution (Single) - resolution of the image in DPI
    '                              (dots per inch)
    '   - tColourType (UDT) - colour depth of the imaged - monochromatic (BW),
    '                         colours of grey (GREY), full colours (COLOUR)
    '   - sngImageLeft, sngImageTop, sngImageRight, sngImageBottom (Single) -
    '       values determine the rectangle on the scanner glass that will
    '       be scanned (default units are inches) - if you set Right and Bottom
    '       values to 0, the module sets maximum values the scanner driver allows
    '       (the bottom right corner of the scanner glass)
    '   - sBMPFileName (String) - the file name of the saved image
    '
    ' Function returns 0 if OK, 1 if an error occurs
    '----------------------------------------------------------------------------

    DIM lRtn AS LONG
    DIM lTmp AS LONG
    DIM blTwainOpen AS INTEGER
    DIM lhDIB AS LONG

    ON ERROR GOTO ErrPlace

    '-------------------------------
    ' Open Twain Data Source Manager
    '-------------------------------
    lRtn = OpenTwainDSM()
    IF lRtn THEN GOTO ErrPlace
    blTwainOpen = %True

    '-----------------------
    ' Open Twain Data Source
    '-----------------------
    lRtn = OpenTwainDS()
    IF lRtn THEN GOTO ErrPlace

    '-----------------------------------------------------------
    ' Set all important attributes of the image and the transfer
    '-----------------------------------------------------------

    '----------------------------------------------------------------------
    ' Set image size and position
    ' If sngImageRight or sngImageBottom is 0 put physical width and height
    '   of the scanner into these values
    '----------------------------------------------------------------------
    IF (sngImageRight = 0) OR (sngImageBottom = 0) THEN
        DIM v AS VARIANT: v = sngImageRight
        lRtn = TwainGetOneValue(%ICAP_PHYSICALWIDTH, v)
        sngImageRight = VARIANT#(v)
        IF lRtn THEN GOTO ErrPlace
        v = sngImageBottom
        lRtn = TwainGetOneValue(%ICAP_PHYSICALHEIGHT, v)
        sngImageBottom = VARIANT#(v)
        IF lRtn THEN GOTO ErrPlace
    END IF

    lRtn = SetImageSize(sngImageLeft, sngImageTop, sngImageRight, sngImageBottom)
    IF lRtn THEN GOTO ErrPlace


    '-----------------------------------------------
    ' Set the image resolution in DPI - both X and Y
    '-----------------------------------------------
    v = sngResolution
    lRtn = TwainSetOneValue(%XRESOLUTION, %FIX32, v): sngResolution = VARIANT#(v)
    IF lRtn THEN GOTO ErrPlace

    v = sngResolution
    lRtn = TwainSetOneValue(%YRESOLUTION, %FIX32, v): sngResolution = VARIANT#(v)
    IF lRtn THEN GOTO ErrPlace

    '--------------------------
    ' Set the image colour type
    '--------------------------
    v = tColourType
    lRtn = TwainSetOneValue(%PIXELTYPE, %UINT16, v): tColourType = VARIANT#(v)
    IF lRtn THEN GOTO ErrPlace

    '----------------------------------------------------------------
    ' If the colour type is fullcolour, set the bitdepth of the image
    '   - 24 bits, 32 bits, ...
    '----------------------------------------------------------------
    v = 24
    IF tColourType = %TWPT_RGB THEN lRtn = TwainSetOneValue(%BITDEPTH, %UINT16, v)

    '---------------------------------------------------
    ' Set number of images you want to transfer (just 1)
    '---------------------------------------------------
    v = 1
    lRtn = TwainSetOneValue(%XFERCOUNT, %INT16, v)
    IF lRtn THEN GOTO ErrPlace

    '----------------------------------------------------
    ' TRANSFER the image with UI disabled.
    '   If successful, lhDIB is filled with handle to DIB
    '----------------------------------------------------
    lRtn = TwainTransfer(%FALSE, lhDIB)
    IF lRtn THEN GOTO ErrPlace

    '------------------
    ' Close Data Source
    '------------------
    lRtn = CloseTwainDS()
    IF lRtn THEN GOTO ErrPlace

    '--------------------------
    ' Close Data Source Manager
    '--------------------------
    lRtn = CloseTwainDSM()
    IF lRtn THEN GOTO ErrPlace
    blTwainOpen = %FALSE

    '----------------------------------
    ' Save DIB handle into the BMP file
    '----------------------------------
    lRtn = SaveDIBToFile(lhDIB, sBMPFileName)
    IF lRtn THEN GOTO ErrPlace

    TransferWithoutUI = 0
    EXIT FUNCTION

ErrPlace:
    IF lhDIB THEN lRtn = GlobalFree(lhDIB)
    IF blTwainOpen THEN lRtn = CloseTwainDS(): lRtn = CloseTwainDSM()
    TransferWithoutUI = 1
END FUNCTION

FUNCTION TransferWithUI(BYVAL sBMPFileName AS STRING) AS LONG

    '-------------------------------------------------------------------
    ' Function transfers one image from Twain data source using the data
    '   source user interface to set image attributes.
    '
    ' Input values
    '   - sBMPFileName (String) - the file name of the saved image
    '
    ' Function returns 0 if OK, 1 if an error occurs
    '-------------------------------------------------------------

    DIM lRtn AS LONG
    DIM blTwainOpen AS INTEGER
    DIM lhDIB AS LONG

    ON ERROR GOTO ErrPlace

    '-------------------------------
    ' Open Twain Data Source Manager
    '-------------------------------
    lRtn = OpenTwainDSM()
    IF lRtn THEN GOTO ErrPlace
    blTwainOpen = %True

    '-----------------------
    ' Open Twain Data Source
    '-----------------------
    lRtn = OpenTwainDS()
    IF lRtn THEN GOTO ErrPlace

    '----------------------------------------------------
    ' TRANSFER the image with UI enabled.
    '   If successful, lhDIB is filled with handle to DIB
    '----------------------------------------------------
    lRtn = TwainTransfer(%True, lhDIB)
    IF lRtn THEN GOTO ErrPlace

    '------------------
    ' Close Data Source
    '------------------
    lRtn = CloseTwainDS()
    IF lRtn THEN GOTO ErrPlace

    '--------------------------
    ' Close Data Source Manager
    '--------------------------
    lRtn = CloseTwainDSM()
    IF lRtn THEN GOTO ErrPlace
    blTwainOpen = %FALSE

    '----------------------------------
    ' Save DIB handle into the BMP file
    '----------------------------------
    lRtn = SaveDIBToFile(lhDIB, sBMPFileName)
    IF lRtn THEN GOTO ErrPlace

    TransferWithUI = 0
    EXIT FUNCTION

ErrPlace:
    IF lhDIB THEN lRtn = GlobalFree(lhDIB)
    IF blTwainOpen THEN lRtn = CloseTwainDS(): lRtn = CloseTwainDSM()
    TransferWithUI = 1
END FUNCTION

FUNCTION PopupSelectSourceDialog() AS LONG

    '------------------------------------------------------------------
    ' Function shows the Twain dialog for selecting default data source
    '
    ' Function returns 0 if OK, 1 if an error occurs
    '------------------------------------------------------------------

    DIM iRtn AS INTEGER
    DIM lRtn AS LONG

    ON ERROR GOTO ErrPlace

    '-------------------------------
    ' Open Twain Data Source Manager
    '-------------------------------
    lRtn = OpenTwainDSM()
    IF lRtn THEN GOTO ErrPlace

    '----------------------------------------------------
    ' Popup "Select source" dialog
    '   DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT
    '----------------------------------------------------
    iRtn = DSM_Entry(m_tAppID, BYVAL 0&, %DG_CONTROL, %DAT_IDENTITY, _
                     %MSG_USERSELECT, m_tSrcID)
    IF iRtn <> %TWRC_SUCCESS THEN
        lRtn = CloseTwainDSM()
        GOTO ErrPlace
    END IF

    '--------------------------------
    ' Close Twain Data Source Manager
    '--------------------------------
    lRtn = CloseTwainDSM()
    IF lRtn THEN GOTO ErrPlace

    PopupSelectSourceDialog = 0
    EXIT FUNCTION

ErrPlace:
    PopupSelectSourceDialog = 1
END FUNCTION

FUNCTION OpenTwainDSM() AS LONG

    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '----------------------------------------------------
    ' Create window that will receive all TWAIN messages
    ' Message loop can be found in TwainTransfer function
    '----------------------------------------------------
    m_lHndMsgWin = CreateWindowEx(0&, "#32770", "TWAIN_MSG_WINDOW", 0&, 10&, 10&, 150&, 50&, 0&, BYVAL %NULL, 0&, BYVAL %NULL)
    IF m_lHndMsgWin = 0 THEN GOTO ErrPlace

    '------------------------------------------------------------
    ' Introduce yourself to TWAIN - MajorNum, MinorNum, Language,
    ' Country, Manufacturer, ProductFamily, ProductName, etc.
    '------------------------------------------------------------
    CALL ZeroMemory(VARPTR(m_tAppID), LEN(m_tAppID))

    m_tAppID.Version.MajorNum = 1
    m_tAppID.Version.Language = %TWLG_CZECH
    m_tAppID.Version.Country = %TWCY_CZECHOSLOVAKIA
    m_tAppID.ProtocolMajor = %TWON_PROTOCOLMAJOR
    m_tAppID.ProtocolMinor = %TWON_PROTOCOLMINOR
    m_tAppID.SupportedGroups1 = %DG_CONTROL OR %DG_IMAGE

    LOCAL UserName, Pbm, pbmTw AS STRING

    UserName = "PowerUser"
    Pbm      = "PowerBASIC Module"
    pbmtw    = "PowerBASIC Module for TWAIN"

    CALL CopyMemory(VARPTR(m_tAppID.Manufacturer(1)), _
                    STRPTR(UserName), _
                    LEN(UserName))
    CALL CopyMemory(VARPTR(m_tAppID.ProductFamily(1)), _
                    STRPTR(Pbm), _
                    LEN(Pbm))
    CALL CopyMemory(VARPTR(m_tAppID.ProductName(1)), _
                    STRPTR(pbmTw), _
                    LEN(pbmTw))

    '--------------------------------------
    ' Open Data Source Manager
    '   DG_CONTROL, DAT_PARENT, MSG_OPENDSM
    '--------------------------------------
    iRtn = DSM_Entry(m_tAppID, BYVAL 0&, %DG_CONTROL, %DAT_PARENT, %MSG_OPENDSM, _
                     m_lHndMsgWin)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    OpenTwainDSM = 0
    EXIT FUNCTION

ErrPlace:
    OpenTwainDSM = 1
END FUNCTION

FUNCTION OpenTwainDS() AS LONG

    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '----------------------------------------------------------------------
    ' Open Data Source
    '   DG_CONTROL, DAT_IDENTITY, MSG_OPENDS
    '
    ' The default data source is opened. If you want user to select the new
    '   default one, call public function PopupSelectSourceDialog.
    '----------------------------------------------------------------------
    CALL ZeroMemory(VARPTR(m_tSrcID), LEN(m_tSrcID))
    iRtn = DSM_Entry(m_tAppID, BYVAL 0&, %DG_CONTROL, %DAT_IDENTITY, %MSG_OPENDS, _
                     m_tSrcID)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    OpenTwainDS = 0
    EXIT FUNCTION

ErrPlace:
    OpenTwainDS = 1
END FUNCTION

FUNCTION CloseTwainDS() AS LONG

    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '----------------------------------------
    ' Close Data Source
    '   DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS
    '----------------------------------------
    iRtn = DSM_Entry(m_tAppID, BYVAL 0&, %DG_CONTROL, %DAT_IDENTITY, _
                     %MSG_CLOSEDS, m_tSrcID)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    CloseTwainDS = 0
    EXIT FUNCTION

ErrPlace:
    CloseTwainDS = 1
END FUNCTION

FUNCTION CloseTwainDSM() AS LONG

    DIM lRtn AS LONG
    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '---------------------------------------
    ' Close Data Source Manager
    '   DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM
    '---------------------------------------
    iRtn = DSM_Entry(m_tAppID, BYVAL 0&, %DG_CONTROL, %DAT_PARENT, %MSG_CLOSEDSM, _
                     m_lHndMsgWin)
    IF iRtn <> %TWRC_SUCCESS THEN
        lRtn = DestroyWindow(m_lHndMsgWin)
        GOTO ErrPlace
    END IF

    '---------------------------
    ' Destroy the message window
    '---------------------------
    lRtn = DestroyWindow(m_lHndMsgWin)
    IF lRtn = 0 THEN GOTO ErrPlace

    CloseTwainDSM = 0
    EXIT FUNCTION

ErrPlace:
    CloseTwainDSM = 1
END FUNCTION

FUNCTION SetImageSize(BYREF sngLeft AS SINGLE, _
                              BYREF sngTop AS SINGLE, _
                              BYREF sngRight AS SINGLE, _
                              BYREF sngBottom AS SINGLE) AS LONG

    DIM tImageLayout AS TW_IMAGELAYOUT
    DIM lRtn AS LONG
    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '-------------------------------------------------------------------
    ' Set the size of the image - in default units
    '   DG_IMAGE, DAT_IMAGELAYOUT, MSG_SET
    '
    ' If you do not select any units the INCHES are selected as default.
    ' The values of Single type are converted into TWAIN TW_FIX32.
    '-------------------------------------------------------------------
    lRtn = FloatToFix32(sngLeft, tImageLayout.Frame.Left)
    IF lRtn THEN GOTO ErrPlace

    lRtn = FloatToFix32(sngTop, tImageLayout.Frame.Top)
    IF lRtn THEN GOTO ErrPlace

    lRtn = FloatToFix32(sngRight, tImageLayout.Frame.Right)
    IF lRtn THEN GOTO ErrPlace

    lRtn = FloatToFix32(sngBottom, tImageLayout.Frame.Bottom)
    IF lRtn THEN GOTO ErrPlace

    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_IMAGE, %DAT_IMAGELAYOUT, %MSG_SET, _
                     tImageLayout)
    IF (iRtn <> %TWRC_SUCCESS) AND (iRtn <> %TWRC_CHECKSTATUS) THEN GOTO ErrPlace

    SetImageSize = 0
    EXIT FUNCTION

ErrPlace:
    SetImageSize = 1
END FUNCTION

FUNCTION TwainTransfer(BYREF blShowUI AS INTEGER, _
                       BYREF lDIBHandle AS LONG) AS LONG

    DIM tUI AS TW_USERINTERFACE
    DIM tPending AS TW_PENDINGXFERS
    DIM lhDIB AS LONG
    DIM tEvent AS TW_EVENT
    DIM tMSG AS MSG
    DIM lRtn AS LONG
    DIM iRtn AS INTEGER

    ON ERROR GOTO ErrPlace

    '---------------------------------------------
    ' Set tUI.ShowUI to 1 (show UI) or 0 (hide UI)
    '---------------------------------------------
    tUI.ShowUI = IIF(blShowUI = %True, 1, 0)
    tUI.ModalUI = 1
    tUI.hParent = m_lHndMsgWin

    '----------------------------------------------
    ' Enable Data Source User Interface
    '   DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS
    '----------------------------------------------
    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_USERINTERFACE, _
                     %MSG_ENABLEDS, tUI)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    '-----------------------------------------------------------------
    ' Process events in the message loop
    '   DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT
    '
    ' There are two messages we are interested in in this message loop
    '   - MSG_XFERREADY - the data source is ready to transfer
    '   - MSG_CLOSEDSREQ - the data source requests to close itself
    '-----------------------------------------------------------------
    WHILE GetMessage(tMSG, 0&, 0&, 0&)
        CALL ZeroMemory(VARPTR(tEvent), LEN(tEvent))
        tEvent.pEvent = VARPTR(tMSG)
        iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_EVENT, _
                         %MSG_PROCESSEVENT, tEvent)
        SELECT CASE tEvent.TWMessage
            CASE %MSG_XFERREADY
                GOTO MSGGET
            CASE %MSG_CLOSEDSREQ
                GOTO MSGDISABLEDS
        END SELECT
        lRtn = TranslateMessage(tMSG)
        lRtn = DispatchMessage(tMSG)
    WEND

MSGGET:
    '----------------------------------------------------
    ' Start transfer
    '   DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET
    '
    ' If transfer is successful you get the handle to DIB
    '----------------------------------------------------
    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_IMAGE, %DAT_IMAGENATIVEXFER, _
                     %MSG_GET, lhDIB)
    IF iRtn <> %TWRC_XFERDONE THEN
        iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_PENDINGXFERS, _
                         %MSG_ENDXFER, tPending)
        iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_USERINTERFACE, _
                         %MSG_DISABLEDS, tUI)
        GOTO ErrPlace
    END IF

    '--------------------------------------------
    ' End transfer
    '   DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER
    '--------------------------------------------
    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_PENDINGXFERS, _
                     %MSG_ENDXFER, tPending)
    IF iRtn <> %TWRC_SUCCESS THEN
        iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_USERINTERFACE, _
                         %MSG_DISABLEDS, tUI)
        GOTO ErrPlace
    END IF

MSGDISABLEDS:
    '-----------------------------------------------
    ' Disable Data Source
    '   DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS
    '-----------------------------------------------
    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_USERINTERFACE, _
                     %MSG_DISABLEDS, tUI)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    lDIBHandle = lhDIB
    TwainTransfer = 0
    EXIT FUNCTION

ErrPlace:
    IF lhDIB THEN lRtn = GlobalFree(lhDIB)
    lDIBHandle = 0
    TwainTransfer = 1
END FUNCTION

FUNCTION SaveDIBToFile(BYREF lhDIB AS LONG, _
                               BYREF sFileName AS STRING) AS LONG

    '---------------------------------------------------------------------------
    ' Function saves the handle to DIB (device independent bitmap) into BMP file
    '---------------------------------------------------------------------------

    DIM tBFH AS BITMAPFILEHEADER
    DIM tBIH AS BITMAPINFOHEADER
    DIM tRGB AS RGBQUAD
    DIM lpDIB AS LONG
    DIM lDIBSize AS LONG
    DIM bDIBits() AS BYTE
    DIM iFileNum AS INTEGER
    DIM lRtn AS LONG

    ON ERROR GOTO ErrPlace

    IF sFileName = "" THEN GOTO ErrPlace

    IF DIR$(sFileName) <> "" THEN
        SETATTR sFileName, %NORMAL
        KILL sFileName
    END IF

    lpDIB = GlobalLock(lhDIB)
    IF lpDIB = 0 THEN GOTO ErrPlace

    CALL CopyMemory(VARPTR(tBIH), lpDIB, LEN(tBIH))

    lDIBSize = LEN(tBIH) + (tBIH.biClrUsed * LEN(tRGB)) + _
               (((tBIH.biWidth * tBIH.biBitCount + 31) \ 32) * 4 * tBIH.biHeight)
    REDIM bDIBits(1 TO lDIBSize) AS BYTE
    CALL CopyMemory(VARPTR(bDIBits(1)), lpDIB, lDIBSize)

    lRtn = GlobalUnlock(lhDIB)
    lRtn = GlobalFree(lhDIB)
    lhDIB = 0

    tBFH.bfType = 19778     ' "BM"
    tBFH.bfSize = LEN(tBFH) + lDIBSize
    tBFH.bfOffBits = LEN(tBFH) + LEN(tBIH) + (tBIH.biClrUsed * LEN(tRGB))

    iFileNum = FREEFILE
    OPEN sFileName FOR BINARY AS #iFileNum
        PUT #iFileNum, , tBFH
        PUT #iFileNum, , bDIBits()
    CLOSE #iFileNum

    SaveDIBToFile = 0
    EXIT FUNCTION

ErrPlace:
    lRtn = GlobalUnlock(lhDIB)
    lRtn = GlobalFree(lhDIB)
    lhDIB = 0
    SaveDIBToFile = 1
END FUNCTION

FUNCTION TwainSetOneValue(BYVAL Cap AS LONG, _
                                  BYVAL ItemType AS LONG, _
                                  BYREF Item AS VARIANT) AS LONG

    '-----------------------------------------------------------------------
    ' There are four types of containers that TWAIN defines for capabilities
    ' (TW_ONEVALUE, TW_ARRAY, TW_RANGE and TW_ENUMERATION)
    ' This module deals with one of them only - TW_ONEVALUE (single value)
    ' To set some capability you have to fill TW_ONEVALUE fields and use
    '   the triplet DG_CONTROL DAT_CAPABILITY MSG_SET
    ' The macros that convert some data types are used here as well
    '-----------------------------------------------------------------------
    ON ERROR GOTO ErrPlace

    DIM tCapability AS TW_CAPABILITY
    DIM tOneValue AS TW_ONEVALUE
    DIM lhOneValue AS LONG
    DIM lpOneValue AS LONG
    DIM lRtn AS LONG
    DIM iRtn AS INTEGER
    DIM tFix32 AS TW_FIX32
    DIM iTmp AS INTEGER

    tCapability.ConType = %TWON_ONEVALUE
    tCapability.Cap = Cap

    tOneValue.ItemType = ItemType

    SELECT CASE ItemType
        CASE %INT16
            tOneValue.Item1 = VARIANT#(Item)
        CASE %UINT16, %BOOL
            IF ToUnsignedShort(VARIANT#(Item), iTmp) THEN GOTO ErrPlace
            CALL CopyMemory(VARPTR(tOneValue.Item1), VARPTR(iTmp), 2&)
        CASE %FIX32
            IF FloatToFix32(VARIANT#(Item), tFix32) THEN GOTO ErrPlace
            CALL CopyMemory(VARPTR(tOneValue.Item1), VARPTR(tFix32), 4&)
    END SELECT

    lhOneValue = GlobalAlloc(%GHND, LEN(tOneValue))
    lpOneValue = GlobalLock(lhOneValue)
    CALL CopyMemory(lpOneValue, VARPTR(tOneValue), LEN(tOneValue))
    lRtn = GlobalUnlock(lhOneValue)
    tCapability.hContainer = lhOneValue

    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_CAPABILITY, %MSG_SET, _
                     tCapability)
    IF iRtn <> %TWRC_SUCCESS THEN
        lRtn = GlobalFree(lhOneValue)
        GOTO ErrPlace
    END IF
    lRtn = GlobalFree(lhOneValue)

    TwainSetOneValue = 0
    EXIT FUNCTION

ErrPlace:
    TwainSetOneValue = 1
END FUNCTION

FUNCTION TwainGetOneValue(BYVAL Cap AS LONG, _
                                  BYREF Item AS VARIANT) AS LONG

    '-----------------------------------------------------------------------
    ' There are four types of containers that TWAIN defines for capabilities
    ' (TW_ONEVALUE, TW_ARRAY, TW_RANGE and TW_ENUMERATION)
    ' This module deals with one of them only - TW_ONEVALUE (single value)
    ' To get some capability you have to fill TW_ONEVALUE fields and use
    '   the triplet DG_CONTROL DAT_CAPABILITY MSG_GET
    ' The macros that convert some data types are used here as well
    '-----------------------------------------------------------------------

    ON ERROR GOTO ErrPlace

    DIM tCapability AS TW_CAPABILITY
    DIM tOneValue AS TW_ONEVALUE
    DIM tFix32 AS TW_FIX32
    DIM lpOneValue AS LONG
    DIM lRtn AS LONG
    DIM iRtn AS INTEGER

    tCapability.ConType = %TWON_ONEVALUE
    tCapability.Cap = Cap

    iRtn = DSM_Entry(m_tAppID, m_tSrcID, %DG_CONTROL, %DAT_CAPABILITY, %MSG_GET, _
                     tCapability)
    IF iRtn <> %TWRC_SUCCESS THEN GOTO ErrPlace

    lpOneValue = GlobalLock(tCapability.hContainer)
    CALL CopyMemory(VARPTR(tOneValue), lpOneValue, LEN(tOneValue))
    lRtn = GlobalUnlock(tCapability.hContainer)
    lRtn = GlobalFree(tCapability.hContainer)

    SELECT CASE tOneValue.ItemType
        CASE %INT16
            Item = tOneValue.Item1
        CASE %UINT16, %BOOL
            Item = FromUnsignedShort(tOneValue.Item1)
        CASE %FIX32
            CALL CopyMemory(VARPTR(tFix32), VARPTR(tOneValue.Item1), 4&)
            Item = Fix32ToFloat(tFix32)
    END SELECT

    TwainGetOneValue = 0
    EXIT FUNCTION

ErrPlace:
    TwainGetOneValue = 1
END FUNCTION

FUNCTION ToUnsignedShort(BYREF lSrc AS LONG, _
                                 BYREF iDst AS INTEGER) AS LONG

    '------------------------------------------------------------------------
    ' Sets number ranging from 0 to 65535 into 2-byte VB Integer
    ' (useful for communicating with other dll that uses unsigned data types)
    '
    ' Function returns 0 is OK, 1 if an error occurs
    '------------------------------------------------------------------------

    ON ERROR GOTO ErrPlace

    IF (lSrc < 0) OR (lSrc > 65535) THEN GOTO ErrPlace

    CALL CopyMemory(VARPTR(iDst), VARPTR(lSrc), 2&)

    ' Another way
    'iDst = IIf(lSrc > 32767, lSrc - 65536, lSrc)

    ToUnsignedShort = 0
    EXIT FUNCTION

ErrPlace:
    ToUnsignedShort = 1
END FUNCTION

FUNCTION FromUnsignedShort(BYREF iSrc AS INTEGER) AS LONG

    '------------------------------------------------------------------------
    ' Gets the 2-byte unsigned number from VB Integer data type
    ' (useful for communicating with other dll that uses unsigned data types)
    '
    ' Function returns unsigned 2-byte value (in VB Long type)
    '------------------------------------------------------------------------

    DIM lTmp AS LONG

    CALL CopyMemory(VARPTR(lTmp), VARPTR(iSrc), 2&)

    ' Another way
    'lTmp = IIf(iSrc < 0, iSrc + 65536, iSrc)

    FromUnsignedShort = lTmp

END FUNCTION

FUNCTION ToUnsignedLong(BYREF sngSrc AS SINGLE, _
                                BYREF lDst AS LONG) AS LONG

    '------------------------------------------------------------------------
    ' Sets number ranging from 0 to 4294967295 into 4-byte VB Long
    ' (useful for communicating with other dll that uses unsigned data types)
    '
    ' Function returns 0 is OK, 1 if an error occurs
    '------------------------------------------------------------------------

    ON ERROR GOTO ErrPlace

    IF (sngSrc < 0) OR (sngSrc > 4294967295#) THEN GOTO ErrPlace

    lDst = IIF(sngSrc > 2147483647, sngSrc - 4294967296#, sngSrc)

    ToUnsignedLong = 0
    EXIT FUNCTION

ErrPlace:
    ToUnsignedLong = 1
END FUNCTION

FUNCTION FromUnsignedLong(BYREF lSrc AS LONG) AS SINGLE

    '------------------------------------------------------------------------
    ' Gets the 4-byte unsigned number from VB Long data type
    ' (useful for communicating with other dll that uses unsigned data types)
    '
    ' Function returns unsigned 4-byte value (in VB Single type)
    '------------------------------------------------------------------------

    DIM sngTmp AS SINGLE

    sngTmp = IIF(lSrc < 0, lSrc + 4294967296#, lSrc)
    FromUnsignedLong = sngTmp

END FUNCTION

FUNCTION Fix32ToFloat(BYREF tFix32 AS TW_FIX32) AS SINGLE

    '----------------------------------------------------------------
    ' Converts TWAIN TW_FIX32 data structure into VB Single data type
    ' (needed for communicating with TWAIN)
    '
    ' Function returns floating-point number in VB Single data type
    '----------------------------------------------------------------

    DIM sngTmp AS SINGLE

    sngTmp = tFix32.Whole + CSNG(FromUnsignedShort(tFix32.Frac) / 65536)
    Fix32ToFloat = sngTmp

END FUNCTION

FUNCTION FloatToFix32(BYREF sngSrc AS SINGLE, _
                              BYREF tFix32 AS TW_FIX32) AS LONG

    '----------------------------------------------------------------
    ' Converts VB Single data type into TWAIN TW_FIX32 data structure
    ' (needed for communicating with TWAIN)
    '
    ' Function returns 0 is OK, 1 if an error occurs
    '----------------------------------------------------------------

    ON ERROR GOTO ErrPlace

    tFix32.Whole = CINT(FIX(sngSrc))
    CALL ToUnsignedShort(CLNG(sngSrc * 65536) AND 65535, tFix32.Frac)
    FloatToFix32 = 0
    EXIT FUNCTION

ErrPlace:
    FloatToFix32 = 1
END FUNCTION
